import itertools
from math import inf

from ..text_model import SeparatedModeText, GroupedText
from ..util import grouping
from ...model import Version


def _opt_c_hc(version: Version, separated: SeparatedModeText) -> GroupedText:
    if len(separated) <= 1:
        return GroupedText([separated])  # 切るところがない場合

    best_grouped: GroupedText = ...
    best_bit_len = inf

    for left, right in itertools.combinations(range(len(separated)), 2):
        grouped = GroupedText(filter(
            lambda x: len(x) != 0, [separated[:left], separated[left:right], separated[right:]]
        ))
        bit_len = grouped.get_segment_length(version)
        if bit_len < best_bit_len:  # todo 等号のありなしを検討
            best_grouped = grouped
            best_bit_len = bit_len

    no_cut_bit_len = separated.get_segment_length(version)
    if no_cut_bit_len <= best_bit_len:  # todo 等号のありなしを検討
        return GroupedText([separated])  # 切らない方が良い場合

    return GroupedText(itertools.chain.from_iterable([_opt_c_hc(version, txt) for txt in best_grouped]))


def optimize_hill_climbing(version: Version, text: str) -> GroupedText:
    return _opt_c_hc(version, SeparatedModeText(grouping(text)))



"""
それぞれのパーツを結合してマイクロQRコードを表す行列を生成するプログラム
"""

from logging import getLogger

from PIL import Image

from ..binary import BinaryMatrix, merge_matrix, empty_matrix, toggle_matrix, BinaryArray
from ..matrix import segment2matrix, get_optimal_mask, get_format_information_matrix, get_function_pattern_matrix
from ..model import Version, ErrorCorrectionLevel as ECL
from ..optimization import analyze_text

logger = getLogger(__name__)


# region セグメントから画像まで
def add_quiet_zone(code: BinaryMatrix, size: int) -> BinaryMatrix:
    """
    クワイエットゾーンを追加

    :param code: コード
    :param size: クワイエットゾーンの幅 (2以上を推奨)
    :return: クワイエットゾーンを追加したコード
    """
    if size < 0:
        raise ValueError('size must be greater than or equal to 0', size)

    h, w = code.shape
    mat = empty_matrix((h + 2 * size, w + 2 * size))
    mat[size:size + h, size:size + w] = code
    return mat


def segment2symbol_matrix(version: Version, ecl: ECL, segment: BinaryArray) -> BinaryMatrix:
    """
    セグメントからマイクロQRコードの行列を作成

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :param segment: セグメント
    :return: マイクロQRコードの行列
    """
    mat_codeword = segment2matrix(version, ecl, segment)
    mask, mat_mask = get_optimal_mask(mat_codeword)
    mat_fi = get_format_information_matrix(version, ecl, mask)
    mat_fp = get_function_pattern_matrix(version)

    code = merge_matrix([mat_fp, mat_fi, mat_codeword, mat_mask])
    return code


def symbol_matrix2image(matrix: BinaryMatrix, size: int = None, quiet_zone: int = 2) -> Image.Image:
    """
    マイクロQRコードの行列から画像を生成

    :param matrix: 行列
    :param size: 画像の一辺のピクセル数 (省略すると1セルが10ピクセルとなるサイズ)
    :param quiet_zone: クワイエットゾーンの幅 (2以上を推奨)
    :return: 画像
    """
    matrix = add_quiet_zone(matrix, quiet_zone)
    matrix = toggle_matrix(matrix)
    img = Image.fromarray(matrix)
    if size is None:
        size = img.height * 10
    img = img.resize((size, size), resample=Image.Resampling.NEAREST)  # PIL 9.1.0から変更
    return img


# segment2image  # 書いても2行だから難しくはない、引数に変更があった時に大変そう？
# endregion


# region 型番など自動識別
def create_symbol_matrix(text: str, ecl: ECL = ECL.NONE) -> BinaryMatrix:
    """
    テキストからマイクロQRコードの行列を作成

    :param text: テキスト
    :param ecl: 誤り訂正レベル
    :return: マイクロQRコードを表す行列
    """
    version, ecl, segment = analyze_text(text, ecl=ecl)
    return segment2symbol_matrix(version, ecl, segment)


def create_symbol_image(text: str, ecl: ECL = ECL.NONE, size: int = None) -> Image.Image:
    """
    テキストからマイクロQRコードの画像を作成

    :param text: テキスト
    :param ecl: 誤り訂正レベル
    :param size: 画像の一辺の長さ
    :return: マイクロQRコードの画像
    """
    matrix = create_symbol_matrix(text, ecl)
    return symbol_matrix2image(matrix, size, 2)
# endregion


# 型番などを指定して作成する関数は必要？
# モードは自動指定しない方が良い？（8ビットバイトモードがかな漢字を含まないため）
# 型番と誤り訂正レベルとテキストを引数とする？



from enum import Enum
from typing import Callable
from mkmqr.binary import bin2arr, BinaryArray


class Mask(Enum):
    """
    マスク

    P49 (PDF 52) 7.8
    """

    Mask00 = (0, lambda y, x: y % 2 == 0)
    Mask01 = (1, lambda y, x: (y // 2 + x // 3) % 2 == 0)
    Mask10 = (2, lambda y, x: ((x * y) % 2 + (x * y) % 3) % 2 == 0)
    Mask11 = (3, lambda y, x: ((x + y) % 2 + (x * y) % 3) % 2 == 0)

    def __init__(self, mask_pattern_value: int, function: Callable[[int, int], bool]):
        self.mask_pattern_value = mask_pattern_value
        """
        マスクパターン参照子の値
        
        P49 (PDF 52) 表10
        """

        self.function = function
        """
        マスクする範囲を示す関数
        
        P49 (PDF 52) 表10
        """

    @property
    def mask_pattern_reference(self) -> BinaryArray:
        """
        マスクパターン参照子

        P49 (PDF 52) 表10
        """
        return bin2arr(self.mask_pattern_value, 2)



"""
真偽値の行列を扱うためのファイル
"""

from typing import Iterable, Union, Tuple
import numpy as np

BinaryMatrix = np.ndarray
"""バイナリ形式の行列"""

_dtype = bool
"""データ型"""


def bin2mat(binaries: Iterable[int], capacity: int) -> BinaryMatrix:
    """
    自然数のビット列を行列に変換する

    :param binaries: 変換する自然数
    :param capacity: 列方向の容量
    :return: 変換後の行列
    """

    if any((b < 0 for b in binaries)):
        raise ValueError(f'binary must be greater than or equal to 0 (index: {[b < 0 for b in binaries].index(True)})')
    if any((b >> capacity != 0 for b in binaries)):
        raise ValueError(f'overflow (index: {[b >> capacity != 0 for b in binaries].index(True)})')

    return np.array([
        [
            binary & (1 << shift) != 0
            for shift in reversed(range(capacity))
        ]
        for binary in binaries
    ], dtype=_dtype)


def mat2str(matrix: BinaryMatrix, on: str = '＠', off: str = '　', row_sep: str = '\n', col_sep: str = '') -> str:
    """
    行列を文字列に変換する

    :param matrix: 変換する行列
    :param on: ビットが立っていることを表す文字
    :param off: ビットが立っていないことを表す文字
    :param row_sep: 行の区切り文字
    :param col_sep: 列の区切り文字
    :return: 変換後の文字列
    """
    return row_sep.join(
        col_sep.join(on if x else off for x in row)
        for row in matrix
    )


def merge_matrix(matrix: Iterable[BinaryMatrix]) -> BinaryMatrix:
    """
    複数の行列をまとめ上げる

    :param matrix: まとめる行列
    :return: まとめ上げた行列
    """
    return np.logical_xor.reduce(matrix, dtype=_dtype)


def toggle_matrix(matrix: BinaryMatrix) -> BinaryMatrix:
    """
    行列の全ての要素を反転する

    :param matrix: 反転する行列
    :return: 反転した行列
    """
    return np.logical_not(matrix)


def empty_matrix(size: Union[int, Tuple[int, int]]):
    """
    空の行列を作成する

    :param size: 行列の大きさ (値を1つのみ指定した場合は正方行列となる)
    :return: 空の行列
    """
    if isinstance(size, int):
        size = size, size
    return np.zeros(size, dtype=bool)



from typing import Iterable

from .text_model import GroupedText, ModeText
from ..model import Mode, InvalidCharacterError


def char2mode(char: str) -> Mode:
    """
    文字に対応している最も低位のモードを取得する

    :param char: 文字
    :return: 対応するモード
    """
    modes = [
        Mode.Numeric,
        Mode.AlphaNumeric,
        Mode.Kanji,
        Mode.EightBitByte,
    ]
    for mode in modes:
        if mode.is_valid(char):
            return mode
    raise InvalidCharacterError(f'Invalid character', char)


def grouping(text: str) -> GroupedText:
    """
    同じモードが連続する箇所でグループ化する

    :param text: グループ化するテキスト
    :return: グループ化したテキスト
    """

    modes = [char2mode(c) for c in text]
    if len(modes) == 0:
        return GroupedText()

    # 下のループのために番兵を仕込む
    modes += [...]
    text += '$'

    # モードが連続する箇所をまとめる
    result = GroupedText()
    prev = ModeText(modes[0], text[0])
    for mode, char in zip(modes[1:], text[1:]):
        if mode == prev.mode:
            prev.text += char
        else:
            result.append(prev)
            prev = ModeText(mode, char)
    return result


def list2str(lst: Iterable):
    """リストを文字列に変換する"""
    return '[' + ', '.join((str(x) for x in lst)) + ']'



"""
テキストの符号化を行う各種モードについてのファイル
"""


from enum import Enum
from typing import List, Optional, Union

from ..binary import bin2arr, BinaryArray, concat_arr


def _each_slice(lst: str, n: int) -> List[str]:
    return [lst[i:i+n] for i in range(0, len(lst), n)]


class _NumericMode:
    """
    数字モード

    P23 (PDF 26) 7.4.3
    """

    def __init__(self):
        self.label = '数字モード'
        self.mode_indicator_value = 0

    def is_valid(self, text: str) -> bool:
        return all((c in '0123465789' for c in text))

    def encode(self, text: str) -> BinaryArray:
        def f(txt: str):
            txt2bin = {1: 4, 2: 7, 3: 10}
            txt_len = len(txt)
            bin_len = txt2bin[txt_len]
            return bin2arr(int(txt), bin_len)

        return concat_arr([f(txt) for txt in _each_slice(text, 3)])

    def bit_length(self, character_count: int) -> int:
        d = 10 * (character_count // 3)
        r = 0 if character_count % 3 == 0 else\
            4 if character_count % 3 == 1 else\
            7  # if character_count % 3 == 2
        return d + r

    def character_count(self, text: str) -> int:
        return len(text)


class _AlphaNumericMode:
    """
    英数字モード

    P24 (PDF 27) 7.4.4
    """

    def __init__(self):
        self.label = '英数字モード'
        self.mode_indicator_value = 1

        self._table = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:'

    def is_valid(self, text: str) -> bool:
        return all((c in self._table for c in text))

    def encode(self, text: str) -> BinaryArray:
        def f(txt: str):
            indices = [self._table.index(c) for c in txt]
            if len(indices) == 1:
                return bin2arr(indices[0], 6)
            else:  # len(indices) == 2
                return bin2arr(indices[0] * 45 + indices[1], 11)

        return concat_arr([f(txt) for txt in _each_slice(text, 2)])

    def bit_length(self, character_count: int) -> int:
        d = 11 * (character_count // 2)
        r = 6 * (character_count % 2)
        return d + r

    def character_count(self, text: str) -> int:
        return len(text)


class _EightBitByteMode:
    """
    8ビットバイトモード

    P25 (PDF 28) 7.4.5
    """

    def __init__(self):
        self.label = '8ビットバイトモード'
        self.mode_indicator_value = 2

        self.encoding = 'shift-jis'  # JIS X0510 P88(PDF 91)の附属書Hに掲載されたJIS X0201のスーパーセット
        # self.encoding = 'ISO_8859_1'  # JIS X0510 P26(PDF 29)に掲載された文字コード
        # self.encoding = 'utf-8'  # クルクルで読み取り可能

    def is_valid(self, text: str) -> bool:
        try:
            # b = text.encode(self._encoding)
            # return all((0x00 <= x <= 0x7F or 0xA0 <= x <= 0xDF for x in b))
            text.encode(self.encoding)
            return True
        except UnicodeError:
            return False

    def encode(self, text: str) -> BinaryArray:
        return concat_arr([bin2arr(b, 8) for b in text.encode(self.encoding)])

    def bit_length(self, character_count: int) -> int:
        return 8 * character_count
    
    def character_count(self, text: str) -> int:
        return len(text.encode(self.encoding))


class _KanjiMode:
    """
    漢字モード

    P27 (PDF 30) 7.4.6
    """

    def __init__(self):
        self.label = '漢字モード'
        self.mode_indicator_value = 3

        self._encoding = 'Shift-JIS'

    def is_valid(self, text: str) -> bool:
        def check(char: str):
            try:
                # return len(char.encode(self._encoding)) == 2
                x = int.from_bytes(char.encode(self._encoding), 'big')
                high, low = divmod(x, 0x100)
                return (0x8140 <= x <= 0x9FFC or 0xE040 <= x <= 0xEBBF) and low != 0x7F
            except UnicodeError:
                return False

        return all((check(c) for c in text))

    def encode(self, text: str) -> BinaryArray:
        def cvt(char: str):
            x = int.from_bytes(char.encode(self._encoding), 'big')
            if 0x8140 <= x <= 0x9FFC:
                x -= 0x8140
            else:  # elif 0xE040 <= x <= 0xEBBF:
                x -= 0xC140
            high, low = divmod(x, 0x100)
            x = 0xC0 * high + low
            return bin2arr(x, 13)

        return concat_arr([cvt(c) for c in text])

    def bit_length(self, character_count: int) -> int:
        return 13 * character_count
    
    def character_count(self, text: str) -> int:
        return len(text)


class Mode(Enum):
    """
    モード

    P18 (PDF 22) 7.3
    """

    Numeric = _NumericMode()
    """数字モード"""

    AlphaNumeric = _AlphaNumericMode()
    """英数字モード"""

    EightBitByte = _EightBitByteMode()
    """8ビットバイトモード"""

    Kanji = _KanjiMode()
    """漢字モード"""

    @property
    def label(self) -> str:
        """モードの名称"""
        return self.value.label

    @property
    def mode_indicator_value(self) -> int:
        """
        モード指示子の値

        P21 (PDF 24) 表2
        """
        return self.value.mode_indicator_value

    def is_valid(self, text: str) -> bool:
        """符号化可能か？"""
        return self.value.is_valid(text)

    def encode(self, text: str) -> Optional[BinaryArray]:
        """符号化を行う"""
        # if not self.value.is_valid(text):
        #     return None
        return self.value.encode(text)

    def character_count(self, text: str) -> int:
        """マルチバイト文字を考慮した文字数"""
        return self.value.character_count(text)

    def bit_length(self, text_or_character_count: Union[int, str]) -> int:
        """
        符号化後の2進データのビット数
        (実際に符号化できるかは考慮しない)

        :param text_or_character_count: テキストまたは文字数
        """
        # [memo]
        # 符号化の際に文字数指示子がOverCapacityErrorを投げる可能性がある
        # また、符号化ができてもシンボルの容量を超える可能性がある
        # そのため符号化→容量チェックの順に行うとtry-exceptとifの両方が必要となり、フローが複雑となる
        # これを防ぐためにデータ長を見積もるメソッドを用意し、先に容量チェックしてから符号化する
        # (容量を見積もることは文字数指示子が収まることの十分条件である はず)
        if isinstance(text_or_character_count, str):
            character_count = self.character_count(text_or_character_count)
        else:
            character_count = text_or_character_count
        return self.value.bit_length(character_count)


# FIXME そもそもEnumである必要はあるのかを含めて、設計を見直す (特に mode == Mode.Numeric のような判定に影響しないか？)
def set_encoding(encoding: str):
    """
    8ビットバイトモードのエンコーディングを指定する

    :param encoding: 文字エンコーディング
    """
    Mode.EightBitByte.value.encoding = encoding



"""
多項式環 F2[x]を扱うためのファイル
"""

from typing import *


F2Array = int  # bitarray
"""有限体F2の配列とみなすビット列"""


class PolynomialRing:
    """多項式環 F2[x]"""

    def __init__(self, coefficient: F2Array):
        """
        :param coefficient: 多項式の係数 (x^3 + x + 1 なら 0b1011)
        """
        if not isinstance(coefficient, F2Array):
            raise TypeError(f'coefficient must be int, but {type(coefficient)} was given')

        if coefficient < 0:
            raise ValueError('coefficient must be greater than or equal to 0')
        self._coefficient = coefficient

    @property
    def coefficient(self):
        return self._coefficient

    # region 演算
    def __add__(self, other: 'PolynomialRing') -> 'PolynomialRing':
        return PolynomialRing(self._coefficient ^ other._coefficient)

    def __sub__(self, other: 'PolynomialRing') -> 'PolynomialRing':
        return self + other

    def __mul__(self, other: 'PolynomialRing') -> 'PolynomialRing':
        left_cof = self._coefficient
        right_cof = other._coefficient

        # 筆算の要領で足し合わせていく
        result = 0
        while right_cof > 0:
            if right_cof & 1 == 1:  # 末尾の1桁に注目する
                result ^= left_cof  # 有限体F2の加算はビットのXORに相当
            right_cof >>= 1  # 処理した桁を破棄し、次の桁を末尾へ降ろす
            left_cof <<= 1  # 足し合わせるために位取りを合わせる

        return PolynomialRing(result)

    def __truediv__(self, other: 'PolynomialRing') -> 'PolynomialRing':
        return self.__divmod__(other)[0]

    def __mod__(self, other: 'PolynomialRing') -> 'PolynomialRing':
        return self.__divmod__(other)[1]

    def __divmod__(self, other: 'PolynomialRing') -> Tuple['PolynomialRing', 'PolynomialRing']:
        left_cof = self._coefficient
        right_cof = other._coefficient

        if right_cof == 0:
            raise ZeroDivisionError()

        def bin_digit_num(n):
            """二進数の桁数を数える"""
            m = 0
            while n != 0:
                n >>= 1
                m += 1
            return m

        quotient = 0
        while True:
            shift_amount = bin_digit_num(left_cof) - bin_digit_num(right_cof)
            if shift_amount < 0:
                break
            left_cof ^= right_cof << shift_amount
            quotient ^= 1 << shift_amount

        return PolynomialRing(quotient), PolynomialRing(left_cof)
    # endregion

    def __hash__(self) -> int:
        return hash(self._coefficient)

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._coefficient == other._coefficient

    def __str__(self) -> str:
        return f'{self._coefficient:b}'

    def __repr__(self):
        return f'PolynomialRing({bin(self._coefficient)})'



"""
コード語列を表す行列を作成するプログラム
"""

from logging import getLogger

from ..binary import bin2arr, concat_arr, BinaryArray, BinaryMatrix, arr2bin, empty_matrix, arr2str
from ..error_correction import ReedSolomonCode, ResidueFieldOperator, PolynomialRing
from ..model import Version, ErrorCorrectionLevel as ECL, OverCapacityError, values
from ..util import Case

logger = getLogger(__name__)


def padding_to_8bit(arr: BinaryArray) -> BinaryArray:
    """8bitの倍数になるように末尾に0を追加する"""
    n = len(arr) % 8
    n = (8 - n) % 8
    return concat_arr([arr, bin2arr(0, n)])


# region データコード語
def add_terminator(version: Version, arr: BinaryArray, capacity: int) -> BinaryArray:
    """
    配列に終端パターンを追加

    :param version: 型番
    :param arr: 終端パターンを追加する配列
    :param capacity: ビット単位の容量
    :return: 終端パターンを追加した配列
    """
    arr = concat_arr([arr, version.terminator])
    return arr[:capacity]


def add_padding_bit(arr: BinaryArray, capacity: int):
    """
    配列に埋め草ビットを追加

    :param arr: 埋め草ビットを追加する配列
    :param capacity: ビット単位の容量
    :return: 埋め草ビットを追加した配列
    """
    arr = padding_to_8bit(arr)
    return arr[:capacity]


def add_padding_codeword(arr: BinaryArray, capacity: int):
    """
    配列に埋め草コード語を追加

    :param arr: 埋め草コード語を追加する配列
    :param capacity: ビット単位の容量
    :return: 埋め草コード語を追加した配列
    """
    padding_codeword_num, remaining = divmod(capacity - len(arr), 8)  # remainingは 0 or 4
    cw = [
        bin2arr(0b1110_1100, 8) if i % 2 == 0 else bin2arr(0b0001_0001, 8)
        for i in range(padding_codeword_num)
    ]
    return concat_arr([arr] + cw + [bin2arr(0, remaining)])


def segment2data_codeword(version: Version, ecl: ECL, segment: BinaryArray) -> BinaryArray:
    """セグメントをデータコード語に変換"""
    logger.debug(f'[convert segment to data codeword]')

    capacity = values.get_data_bit_capacity(version, ecl)
    if len(segment) > capacity:
        raise OverCapacityError(f'Segment({len(segment)}-bit) is over capacity({capacity}-bit)', segment)

    arr = segment
    logger.debug(f'segment: {len(segment)}/{capacity} bits')
    arr = add_terminator(version, arr, capacity)
    logger.debug(f'add terminator: {len(arr)}/{capacity} bits')
    arr = add_padding_bit(arr, capacity)
    logger.debug(f'add remaining bit: {len(arr)}/{capacity} bits')
    arr = add_padding_codeword(arr, capacity)
    logger.debug(f'add remaining codeword: {len(arr)}/{capacity} bits')

    return arr
# endregion


# region 誤り訂正コード語
def get_generator_polynomial(rf_op: ResidueFieldOperator, degree: int):
    """
    生成多項式を取得

    :param rf_op: 剰余環として扱うためのインスタンス
    :param degree: 生成多項式の次数
    :return: 生成多項式
    """
    def exp2gen(*exp: int):
        return [rf_op.from_exp(e) for e in exp]

    return Case(degree)\
        .when(2, exp2gen(0, 25, 1))\
        .when(5, exp2gen(0, 113, 164, 166, 119, 10))\
        .when(6, exp2gen(0, 166, 0, 134, 5, 176, 15))\
        .when(8, exp2gen(0, 175, 238, 208, 249, 215, 252, 196, 28))\
        .when(10, exp2gen(0, 251, 67, 46, 61, 118, 70, 64, 94, 32, 45))\
        .when(14, exp2gen(0, 199, 249, 155, 48, 190, 124, 218, 137, 216, 87, 207, 59, 22, 91))\
        .value


def setup_rs_code(version: Version, ecl: ECL) -> ReedSolomonCode:
    """RS符号を計算するためのインスタンスを取得"""
    primitive_poly = PolynomialRing(0b1_0001_1101)
    rf_op = ResidueFieldOperator(primitive_poly)
    degree = values.get_error_correction_codeword_num(version, ecl)
    generator_poly = get_generator_polynomial(rf_op, degree)
    return ReedSolomonCode(rf_op, generator_poly)


def get_error_correction_codeword(version: Version, ecl: ECL, data_codeword: BinaryArray) -> BinaryArray:
    """
    データコード語から誤り訂正コード語を取得

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :param data_codeword: 誤り訂正コード語を計算する対象のデータコード語
    :return: 計算した誤り訂正コード語
    """
    data_codeword = padding_to_8bit(data_codeword)  # 末尾が4bitでも8bitに合わせる
    poly = [PolynomialRing(arr2bin(data_codeword[i:i+8])) for i in range(0, len(data_codeword), 8)]  # 8bitごとに区切る

    rs_code = setup_rs_code(version, ecl)
    ecc = rs_code.encode(poly)
    return concat_arr([bin2arr(e.coefficient, 8) for e in ecc])
# endregion


# region 行列
def place_codeword(version: Version, codeword: BinaryArray) -> BinaryMatrix:
    """
    コード語列を行列に配置

    :param version: 型番
    :param codeword: コード語列
    :return: コード語列を配置した行列
    """
    n = version.size
    if len(codeword) != (n-1)**2 - 8**2:
        raise ValueError(f'Codewords must be {(n-1)**2 - 8**2}bit, but it is {len(codeword)}bit', codeword)

    def cursor():
        """
        ビットを配置する座標(i,j)を順に返す
        原点は左上、iは下方向、jは右方向を表す
        """
        for idx, j in enumerate(range(n-1, 0, -2)):
            rng = range(9, n) if j <= 8 else range(1, n)  # j<=8のときは切り出しパターンを避けなければならない
            if idx % 2 == 0:  # 最初は上方向(iの減少方向)に配置していく
                rng = reversed(rng)  # 上下に往復するように配置するため、向きを反転する
            for i in rng:  # ジグザグに蛇行して配置していく
                yield i, j
                yield i, j-1

    code = empty_matrix(n)
    for (i, j), c in zip(cursor(), codeword):
        code[i, j] = c
    return code


def segment2matrix(version: Version, ecl: ECL, segment: BinaryArray) -> BinaryMatrix:
    """
    セグメントを行列に変換

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :param segment: セグメント
    :return: セグメントの情報を格納した行列
    """
    data_codeword = segment2data_codeword(version, ecl, segment)
    ec_codeword = get_error_correction_codeword(version, ecl, data_codeword)
    codeword = concat_arr([data_codeword, ec_codeword])
    logger.info('codewords: ' + arr2str(data_codeword, byte_sep=' ') + ', ' + arr2str(ec_codeword, byte_sep=' '))
    mat_codeword = place_codeword(version, codeword)
    return mat_codeword
# endregion



"""
符号化後の2進データが最短になるようにテキストを解析するプログラム
"""


from logging import getLogger
from typing import List, Tuple, Union, Container

from .algorithm import optimize
from .util import char2mode, list2str
from ..binary import BinaryArray, concat_arr, arr2str
from ..model import Version, ErrorCorrectionLevel as ECL, values, OverCapacityError, InvalidPairError

logger = getLogger(__name__)


def text2mixing_segment(version: Version, text: str) -> BinaryArray:
    grouped = optimize(version, text)
    logger.debug(f'({version}, {text}) -> {grouped}')
    return concat_arr([sub.get_segment(version) for sub in grouped])


def text2mixing_segment_length(version: Version, text: str):
    grouped = optimize(version, text)
    return sum((sub.get_segment_length(version) for sub in grouped))


def analyze_text(
        text: str, version: Union[Version, Container[Version]] = ..., ecl: Union[ECL, Container[ECL]] = ...
) -> Tuple[Version, ECL, BinaryArray]:
    """
    テキストを解析して最適な型番、誤り訂正レベル、セグメントを計算する

    :param text: テキスト
    :param version: 最大の型番 (あるいは使用してもよい型番の一覧)
    :param ecl: 必要な誤り訂正レベル (あるいは使用してもよい誤り訂正レベルの一覧)
    :return: 型番, 誤り訂正レベル, セグメント
    """
    # region 前処理  # これ以降は引数のversion,eclを使用しない 一覧を表すversions,eclsを使用する
    from typing import TypeVar
    T = TypeVar('T')

    def parse_arg(arg: Union[T, Container[T]], reference: List[T]) -> List[T]:  # ...の型が上手く表せなかった
        if arg is ...:
            return reference  # そのまま返す
        elif isinstance(arg, Container):
            return [x for x in reference if x in arg]  # 共通部分だけを返す
        else:
            return reference[:reference.index(arg)+1]  # argまでを返す

    versions = parse_arg(version, [Version.M1, Version.M2, Version.M3, Version.M4])
    """探索順を考慮した型番の一覧"""
    ecls = parse_arg(ecl, [ECL.Q, ECL.M, ECL.L, ECL.NONE])
    """探索順を考慮した誤り訂正レベルの一覧"""
    modes = {char2mode(c) for c in text}
    """使用されているモードの一覧"""

    logger.debug(f'versions: {list2str(versions)}')
    logger.debug(f'ecls: {list2str(ecls)}')
    logger.debug(f'modes: {list2str(modes)}')
    # endregion

    # region M1は特殊なので別途処理  # これ以降は_version,_eclを個別の要素を表すために使用する
    _version = Version.M1
    """注目する型番"""
    _ecl = ECL.NONE
    """注目する誤り訂正レベル"""

    if _version in versions and _ecl in ecls:
        if values.check_combination(version=_version, ecl=_ecl, mode=modes):
            seg_len = text2mixing_segment_length(_version, text)
            capacity = values.get_data_bit_capacity(_version, _ecl)
            if seg_len <= capacity:
                segment = text2mixing_segment(_version, text)
                return _version, _ecl, segment
        else:
            logger.debug(f'{_version}: invalid pair')

    # M1と誤り検出のみはこの組み合わせでしか使えないため一覧から除外する
    if _version in versions:
        versions.remove(_version)
    if _ecl in ecls:
        ecls.remove(_ecl)

    # 型番と誤り訂正レベルのいずれかが空になってしまったら終了
    if len(versions) == 0 or len(ecls) == 0:
        raise InvalidPairError('there is no pair that meets the conditions')
    # endregion

    # region 型番の選択  # ここで_versionが確定する
    exists_valid_pair = False  # 容量が不足しているのか、設定が不正なのかを区別するため
    _ecl = ecls[-1]  # 一番容量が多い(=低い)レベルで試す
    logger.debug(f'select version from {list2str(versions)} (lowest ecl: {_ecl})')

    for _version in versions:
        if not values.check_combination(version=_version, ecl=_ecl, mode=modes):
            logger.debug(f'{_version}: invalid pair')
            continue
        exists_valid_pair = True  # 少なくとも1つは有効な組み合わせが存在した

        seg_len = text2mixing_segment_length(_version, text)
        capacity = values.get_data_bit_capacity(_version, _ecl)
        if seg_len > capacity:
            logger.debug(f'{_version}: over capacity')
            continue

        logger.debug(f'{_version}: OK')
        break  # 容量に収まればループを打ち切って型番を確定する (次の工程の番兵を求めるだけなので、セグメントはまだ求めない)
    else:
        # breakされなかった場合 (=条件を満たす組み合わせが存在しなかった場合)
        if exists_valid_pair:
            raise OverCapacityError('there was a pair that met the conditions, but there was not enough capacity')
        else:
            raise InvalidPairError('there is no pair that meets the conditions')
    # endregion

    # region 誤り訂正レベル  # 求めた_versionに_eclをすり合わせる
    logger.debug(f'select ecl from {list2str(ecls)} (version: {_version})')
    for _ecl in ecls:
        if not values.check_combination(version=_version, ecl=_ecl, mode=modes):
            logger.debug(f'{_ecl}: invalid pair')
            continue

        seg_len = text2mixing_segment_length(_version, text)
        capacity = values.get_data_bit_capacity(_version, _ecl)
        if seg_len > capacity:
            logger.debug(f'{_ecl}: over capacity')
            continue

        segment = text2mixing_segment(_version, text)
        logger.debug(f'{_ecl}: OK')
        logger.info(f'analyzed result: {_version}, {_ecl}')
        logger.info(f'binary data: {arr2str(segment)} ({len(segment)} / {capacity} bits)')
        return _version, _ecl, segment

    raise RuntimeError('it never come here')  # 型番の選択結果が番兵となるため、ここには絶対に来ないはず
    # endregion





"""
RS符号を計算するためのファイル
"""

from .polynomial_ring import PolynomialRing
from .residue_field import ResidueFieldOperator
from typing import List, Union


class ReedSolomonCode:
    """RS符号を計算するクラス"""

    def __init__(
            self,
            primitive_polynomial: Union[PolynomialRing, ResidueFieldOperator],
            generator_polynomial: List[PolynomialRing]
    ):
        """
        :param primitive_polynomial: 原始多項式
        :param generator_polynomial: 生成多項式 (高次 gp[0] ← ... → gp[-1] 低次)
        """
        if isinstance(primitive_polynomial, PolynomialRing):
            rf_op = ResidueFieldOperator(primitive_polynomial)
        else:
            rf_op = primitive_polynomial

        self.rf_op = rf_op
        """剰余体の計算補助"""
        self.generator_polynomial = generator_polynomial
        """生成多項式"""

    @property
    def primitive_polynomial(self):
        """原始多項式"""
        return self.rf_op.primitive_polynomial

    def encode(self, data: List[PolynomialRing]) -> List[PolynomialRing]:
        """
        符号化を行う (計算前に自動で桁をシフトする)

        :param data: RS符号を求めるデータ
        :return: 求めたRS符号
        """
        n = len(self.generator_polynomial) - 1
        trail = [PolynomialRing(0) for _ in range(n)]
        return self.encode_raw(data + trail)

    def encode_raw(self, data: List[PolynomialRing]) -> List[PolynomialRing]:
        """
        符号化を行う (計算前に桁をシフトしない)

        :param data: RS符号を求めるデータ
        :return: 求めたRS符号
        """
        temp = data.copy()

        # 筆算のように最高次から順番に引いていく
        for i in range(len(temp) - len(self.generator_polynomial) + 1):
            # 商を求める (最高次は1のはずだが、一応割る)
            q = temp[i] / self.generator_polynomial[0]
            for j in range(len(self.generator_polynomial)):
                temp[i+j] = self.rf_op.sub(temp[i+j], q * self.generator_polynomial[j])

        # 剰余だけ取り出す
        return temp[-(len(self.generator_polynomial)-1):]



from typing import Optional, Protocol, List, Generic, TypeVar

from ..binary import bin2arr, concat_arr, BinaryArray
from ..factory import text2segment
from ..model import Mode, Version, values


def mode2char(mode: Optional[Mode]) -> str:  # todo 設置場所を考える
    """
    各モードを1文字で表す

    :param mode: モード
    :return: 半角1文字
    """
    if mode == Mode.Numeric:
        return 'N'
    elif mode == Mode.AlphaNumeric:
        return 'A'
    elif mode == Mode.EightBitByte:
        return '8'
    elif mode == Mode.Kanji:
        return 'K'
    else:
        return 'x'


class SupportsMode(Protocol):
    """モード取得可能"""

    @property
    def mode(self) -> Mode:
        return ...

    @property
    def text(self) -> str:
        return ...


class SupportsSegment(Protocol):
    """セグメント化可能"""

    def get_segment_length(self, version: Version) -> int:
        return ...

    def get_segment(self, version: Version) -> BinaryArray:
        return ...


class ModeText(SupportsSegment, SupportsMode):
    """モードが付与されたテキスト"""

    def __init__(self, mode: Mode, text: str):
        self._mode = mode
        """符号化のモード"""
        self._text = text
        """テキスト"""

    @property
    def mode(self) -> Mode:
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = value

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value):
        self._text = value

    def __len__(self):
        return len(self.text)

    def __str__(self):
        c = mode2char(self.mode)
        return f'({c}:{self.text})'

    def __hash__(self):
        return hash((self.mode, self.text))

    def __eq__(self, other):
        if not isinstance(other, ModeText):
            return False
        return self.mode == other.mode and self.text == other.text

    def __deepcopy__(self, memo=...):
        return ModeText(self.mode, self.text)

    def get_segment_length(self, version: Version) -> int:
        mi_len = version.mode_indicator_length
        cci_len = values.get_character_count_indicator_length(version, self.mode)
        data_len = self.mode.bit_length(self.text)
        return mi_len + cci_len + data_len

    def get_segment(self, version: Version) -> BinaryArray:
        return text2segment(version, self.mode, self.text)


class SeparatedModeText(List[ModeText], SupportsSegment):
    """
    モードが付与されたテキスト
    内部的に複数のグループに分けられている
    """

    @property
    def mode(self) -> Optional[Mode]:
        modes = {item.mode for item in self}
        if Mode.EightBitByte in modes:
            return Mode.EightBitByte
        elif Mode.Kanji in modes:
            if len(modes - {Mode.Kanji}) > 0:
                return Mode.EightBitByte
            else:
                return Mode.Kanji
        elif Mode.AlphaNumeric in modes:
            return Mode.AlphaNumeric
        elif Mode.Numeric in modes:
            return Mode.Numeric
        else:
            return None  # len(...) == 0

    @property
    def text(self) -> str:
        return ''.join((sub_text.text for sub_text in self))

    def get_segment_length(self, version: Version) -> int:
        if len(self) == 0:
            return 0
        mi_len = version.mode_indicator_length
        cci_len = values.get_character_count_indicator_length(version, self.mode)
        data_len = self.mode.bit_length(self.text)
        return mi_len + cci_len + data_len

    def get_segment(self, version: Version) -> BinaryArray:
        if len(self) == 0:
            return bin2arr(0, 0)
        return text2segment(version, self.mode, self.text)

    def __str__(self) -> str:
        # x = ', '.join((str(sub_text) for sub_text in self))
        # return f'<{mode2char(self.mode)}:{x}>'
        return f'({mode2char(self.mode)}:{self.text})'

    def __getitem__(self, item):
        v = super().__getitem__(item)
        if isinstance(v, list):
            return SeparatedModeText(v)
        else:
            return v


T = TypeVar('T', bound=SupportsSegment)


class GroupedText(Generic[T], List[T], SupportsSegment):
    """セグメント化するグループに分けられたテキスト"""

    def get_segment_length(self, version: Version) -> int:
        return sum((item.get_segment_length(version) for item in self))

    def get_segment(self, version: Version) -> BinaryArray:
        return concat_arr([item.get_segment(version) for item in self])

    def __str__(self) -> str:
        x = ', '.join((str(item) for item in self))
        return f'[{x}]'

    def __getitem__(self, item):
        v = super().__getitem__(item)
        if isinstance(v, list):
            return GroupedText(v)
        else:
            return v



"""
真偽値の配列を扱うためのファイル
"""

from typing import Iterable
import numpy as np

BinaryArray = np.ndarray
"""バイナリ形式の配列"""

_dtype = bool
"""データ型"""


def arr2bin(array: BinaryArray) -> int:
    """
    配列のビット列を自然数に変換する

    :param array: 変換する配列
    :return: 変換後の自然数
    """
    return int(array.dot(2**np.arange(array.size)[::-1]))


def bin2arr(binary: int, capacity: int) -> BinaryArray:
    """
    自然数のビットを配列に変換する

    :param binary: 変換する自然数
    :param capacity: ビットの容量
    :return: 変換後の配列
    """
    if binary < 0:
        raise ValueError(f'binary must be greater than or equal to 0', binary)
    if binary >> capacity != 0:
        raise ValueError(f'overflow ({binary:b} : {capacity} bits)', binary, capacity)

    return np.array([
        binary & (1 << shift) != 0
        for shift in reversed(range(capacity))
    ], dtype=_dtype)


def arr2str(array: BinaryArray, on: str = '1', off: str = '0', byte_sep: str = '') -> str:
    """
    配列を文字列に変換する

    :param array: 変換する配列
    :param on: ビットが立っていることを表す文字
    :param off: ビットが立っていないことを表す文字
    :param byte_sep: 1バイトごとに挿入する区切り文字
    :return: 0,1から成る文字列
    """
    s = ''.join((on if b else off for b in array))
    s = byte_sep.join((s[i:i+8] for i in range(0, len(s), 8)))
    return s


def concat_arr(arrays: Iterable[BinaryArray]):
    """
    配列を連結する

    :param arrays: 連結する配列のリスト
    :return: 連結した配列
    """
    return np.concatenate(arrays, dtype=_dtype)



"""
多項式環を剰余体として扱うためのクラス
"""
from typing import Tuple

from .polynomial_ring import PolynomialRing, F2Array


class ResidueFieldOperator:
    """多項式環を剰余体として扱うためのクラス"""

    def __init__(self, primitive_polynomial: PolynomialRing):
        """
        :param primitive_polynomial: 原始多項式
        """
        self.primitive_polynomial = primitive_polynomial
        """原始多項式"""

    def add(self, left: PolynomialRing, right: PolynomialRing) -> PolynomialRing:
        """
        加算

        :param left: 左オペランド
        :param right: 右オペランド
        :return: 和
        """
        return (left + right) % self.primitive_polynomial

    def sub(self, left: PolynomialRing, right: PolynomialRing) -> PolynomialRing:
        """
        減算

        :param left: 左オペランド
        :param right: 右オペランド
        :return: 差
        """
        return (left - right) % self.primitive_polynomial

    def mul(self, left: PolynomialRing, right: PolynomialRing) -> PolynomialRing:
        """
        乗算

        :param left: 左オペランド
        :param right: 右オペランド
        :return: 積
        """
        return (left * right) % self.primitive_polynomial

    def div(self, left: PolynomialRing, right: PolynomialRing) -> PolynomialRing:
        """
        除算

        :param left: 左オペランド
        :param right: 右オペランド
        :return: 商
        """
        return (left * self.inv(right)) % self.primitive_polynomial

    def inv(self, value: PolynomialRing) -> PolynomialRing:
        """
        逆元

        :param value: オペランド
        :return: 逆元
        """
        if value.coefficient == 0:
            raise ZeroDivisionError()
        _, (inv, _) = self._ext_gcd(value, self.primitive_polynomial)
        return inv

    @classmethod
    def _ext_gcd(cls, a: PolynomialRing, b: PolynomialRing) \
            -> Tuple[PolynomialRing, Tuple[PolynomialRing, PolynomialRing]]:
        """
        | 拡張ユークリッドの互除法
        | a*x + b*y == 1 となる (x, y)を求める

        :param a:
        :param b:
        :return: aとbの最大公約数, (x, y)
        """
        if b.coefficient == 0:
            return a, (PolynomialRing(1), PolynomialRing(0))
        d, (y, x) = cls._ext_gcd(b, a % b)
        y -= a/b * x
        return d, (x, y)

    def from_exp(self, exp: int):
        """
        べき表現から剰余体の元を取得する

        :param exp: べき表現
        :return: 対応する剰余体の元
        """
        return PolynomialRing(1 << exp) % self.primitive_polynomial

    def from_coefficient(self, coefficient: F2Array):
        """
        多項式表現から剰余体の元を取得する

        :param coefficient: 多項式表現
        :return: 対応する剰余体の元
        """
        return PolynomialRing(coefficient) % self.primitive_polynomial



import itertools
from copy import deepcopy
from logging import getLogger
from math import inf
from typing import List

from ..text_model import GroupedText, ModeText
from ..util import list2str, grouping
from ...model import Mode, Version

logger = getLogger(__name__)


def _get_merged_mode(mode1: Mode, mode2: Mode) -> Mode:
    modes = (mode1, mode2)
    if modes == (Mode.Kanji, Mode.Kanji):
        return Mode.Kanji
    elif Mode.Kanji in modes or Mode.EightBitByte in modes:
        return Mode.EightBitByte
    elif Mode.AlphaNumeric in modes:
        return Mode.AlphaNumeric
    else:
        return Mode.Numeric


def _merge_edges(grouped: GroupedText, edges_to_merge: List[bool]):
    """
    指定の境界でグループを結合する (破壊操作)

    :param grouped: グループ化したテキスト (破壊操作)
    :param edges_to_merge: 結合する境界 (破壊操作)
    """
    logger.debug('edges: ' + list2str(edges_to_merge))
    logger.debug('grouped: ' + list2str(grouped))

    edge = 0
    while edge < len(grouped) - 1:  # 長さが変わるためforではなくwhileでループ
        if edges_to_merge[edge]:
            logger.debug(f'- merged at edge {edge}')

            grouped[edge + 1] = ModeText(
                _get_merged_mode(grouped[edge].mode, grouped[edge + 1].mode),
                grouped[edge].text + grouped[edge + 1].text
            )

            # popでインデックスが1つずれるのでedgeはインクリメントしない
            grouped.pop(edge)
            edges_to_merge.pop(edge)
        else:
            logger.debug(f'- not merged at edge {edge}')
            edge += 1  # 次の境界へ

        logger.debug('edges: ' + list2str(edges_to_merge))
        logger.debug('grouped: ' + list2str(grouped))


def optimize_brute_force(version: Version, text: str) -> GroupedText:
    """
    最もビット数が短くなるように指定のテキストをグループ化する

    :param version: 型番
    :param text: テキスト
    :return: ビット数が最短となる区切りのグループ
    """
    # memo グループ数nに対して2**(n-1)回のループを行うため、重くなりやすい 取り扱いに注意
    logger.debug(f'[optimize by brute force]')

    grouped = GroupedText(grouping(text))
    edge_num = len(grouped) - 1

    best_grouped = ...
    best_length = inf

    for edges_to_merge in itertools.product([True, False], repeat=edge_num):
        grp = deepcopy(grouped)  # 元のgroupedを破壊しないようにコピー
        egs = list(edges_to_merge)  # tupleは操作が効かないのでlistへ
        _merge_edges(grp, egs)

        ln = grp.get_segment_length(version)
        if ln < best_length:
            best_grouped = grp
            best_length = ln

        x = '*' if ln == best_length else ' '
        logger.debug(f'{x} result {ln:3}bits {list2str(grp)}')

    logger.info(f'{best_length}bits {list2str(best_grouped)}')
    return best_grouped



"""
マイクロQRコードで使用されるモデルをまとめた内部モジュール
"""

from .version import Version
from .mode import Mode, set_encoding
from .error_correction_level import ErrorCorrectionLevel
from .mask import Mask

from .error import InvalidPairError, InvalidCharacterError, OverCapacityError




"""
シンボルを構成する各パーツを作成する内部モジュール
"""

# 各パーツの生成
from .matrix_codeword import (
    # データコード語
    add_terminator,
    add_padding_bit,
    add_padding_codeword,
    segment2data_codeword,
    # 誤り訂正コード語
    get_generator_polynomial,
    setup_rs_code,
    get_error_correction_codeword,
    # 行列
    place_codeword,
    segment2matrix,
)
from .matrix_format_information import (
    place_format_information,
    get_format_information_matrix,
)
from .matrix_mask import (
    get_mask_matrix,
    calc_mask_score,
    get_optimal_mask,
)
from .matrix_function_pattern import (
    get_function_pattern_matrix,
)



"""
形式情報の行列を生成するプログラム
"""

from logging import getLogger

from ..binary import BinaryArray, BinaryMatrix, empty_matrix, arr2str
from ..model import Version, ErrorCorrectionLevel as ECL, Mask, values

logger = getLogger(__name__)


def place_format_information(version: Version, format_information: BinaryArray) -> BinaryMatrix:
    """形式情報を行列に配置"""
    def idx():
        for i in range(1, 8):
            yield 8, i
        for i in range(8, 0, -1):
            yield i, 8

    mat = empty_matrix(version.size)
    for (i, j), fi in zip(idx(), format_information):
        mat[i, j] = fi
    return mat


def get_format_information_matrix(version: Version, ecl: ECL, mask: Mask) -> BinaryMatrix:
    """形式情報を配置した行列を取得"""
    fi = values.get_format_information(version, ecl, mask)
    logger.info(f'format information: ' + arr2str(fi, byte_sep=' '))
    return place_format_information(version, fi)



from enum import Enum

from mkmqr.binary import bin2arr, BinaryArray


class Version(Enum):
    """
    型番

    P14 (PDF 17) 図11
    """

    M1 = (11, bin2arr(0, 3), 0)
    M2 = (13, bin2arr(0, 5), 1)
    M3 = (15, bin2arr(0, 7), 2)
    M4 = (17, bin2arr(0, 9), 3)

    def __init__(self, size: int, terminator: BinaryArray, mode_indicator_length: int):
        self.size = size
        """
        マイクロQRコードの一辺あたりのモジュール数
        
        P17 (PDF 20) 表1
        """

        self.terminator = terminator
        """
        終端パターン
        
        P30 (PDF 33) 7.4.9, P21 (PDF24) 表2 等
        """

        self.mode_indicator_length = mode_indicator_length
        """
        モード指示子の長さ
        
        P21 (PDF24) 表2 等
        """



from .model import InvalidPairError


class Case:
    """
    if-elif-elseの補助
    Python 3.10からはパターンマッチが使える
    """

    def __init__(self, target):
        """
        :param target: 指定条件
        """
        self.target = target
        """指定条件"""
        self._value = ...

    def when(self, condition, value):
        """
        :param condition: 条件
        :param value: 値
        """
        if value is ...:
            raise ValueError(f'... is not supported')
        if self.target == condition:
            self._value = value
        return self

    @property
    def value(self):
        """条件にマッチした値"""
        if self._value is ...:
            raise InvalidPairError(f'Invalid value ({self.target})')
        return self._value



"""
機能パターンの行列を作成するプログラム
"""

from ..binary import BinaryMatrix, empty_matrix
from ..model import Version


def get_function_pattern_matrix(version: Version) -> BinaryMatrix:
    """機能パターンの行列を取得"""
    ptn = empty_matrix(version.size)
    # タイミングパターン
    ptn[::2, 0] = True
    ptn[0, ::2] = True
    # 機能パターン
    ptn[0:7, 0:7] = True
    ptn[1:6, 1:6] = False
    ptn[2:5, 2:5] = True
    return ptn



"""
テキストを指定のモードでセグメントに変換するプログラム
"""

from logging import getLogger

from ..binary import concat_arr, BinaryArray, arr2str
from ..model import Version, Mode, values

logger = getLogger(__name__)


def data2segment(version: Version, mode: Mode, data: BinaryArray, character_count: int) -> BinaryArray:
    """2進データをセグメントに変換"""
    mi = values.get_mode_indicator(version, mode)
    cci = values.get_character_count_indicator(version, mode, character_count)
    logger.debug(f'segment: (mi: {arr2str(mi)})(cci: {arr2str(cci)})(data: {arr2str(data)})')
    return concat_arr([mi, cci, data])


def text2segment(version: Version, mode: Mode, text: str) -> BinaryArray:
    """テキストをセグメントに変換"""
    data = mode.encode(text)
    cc = mode.character_count(text)
    return data2segment(version, mode, data, cc)



"""
マイクロQRコードを作成するための内部モジュール
"""

# データの解析
from .segment import (
    data2segment,
    text2segment,
)

# パーツの合成
from .symbol import (
    add_quiet_zone,
    segment2symbol_matrix,
    create_symbol_matrix,
    symbol_matrix2image,
    create_symbol_image,
)



"""
マイクロQRコードを作成するライブラリ

参考：
`QRコードドットコム <https://www.qrcode.com/codes/microqr.html>`_
`JIS X0510 <https://www.jisc.go.jp/app/jis/general/GnrJISNumberNameSearchList?show&jisStdNo=X0510>`_
"""

from .__version import __version__

from .binary import (
    BinaryArray,
    BinaryMatrix,
    arr2str,
    mat2str,
    arr2bin,
    bin2arr,
    bin2mat,
    concat_arr,
    merge_matrix,
    toggle_matrix,
    empty_matrix,
)

# from .error_correction import ()

from .model import (
    Version,
    Mode,
    ErrorCorrectionLevel,
    Mask,
    values,
    set_encoding,
    InvalidPairError,
    InvalidCharacterError,
    OverCapacityError,
)

from .matrix import (
    segment2matrix,
    get_format_information_matrix,
    get_optimal_mask,
    get_function_pattern_matrix,
)

from .factory import (
    data2segment,
    text2segment,
    add_quiet_zone,
    segment2symbol_matrix,
    symbol_matrix2image,
    create_symbol_matrix,
    create_symbol_image,
)

from .optimization import (
    analyze_text,
)



from typing import Set, Union

from .error import OverCapacityError, InvalidPairError
from ..binary import concat_arr, bin2arr, BinaryArray
from ..error_correction import PolynomialRing
from ..model import Version, ErrorCorrectionLevel as ECL, Mask, Mode
from ..util import Case


# todo 1つから決まる値はEnumに持たせているが、表記の統一のためここに移すか検討する


# region 組み合わせで決まる値
def _assert_mode(version: Version, mode: Mode) -> None:
    """
    型番とモードの組み合わせが妥当か検証

    :raise InvalidPairError: 不適切な組み合わせが与えられたとき
    """
    if version == Version.M1 and mode != Mode.Numeric:
        raise InvalidPairError('M1 only supports number mode')
    if version == Version.M2 and mode not in [Mode.Numeric, Mode.AlphaNumeric]:
        raise InvalidPairError('M2 only supports number mode and alphanumeric mode')


def get_data_bit_capacity(version: Version, ecl: ECL) -> int:
    """
    データ容量をビット単位で取得

    P31 (PDF 34) 表7

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :return: データ容量(ビット単位)
    """
    return Case([version, ecl]) \
        .when([Version.M1, ECL.NONE], 20) \
        .when([Version.M2, ECL.L], 40) \
        .when([Version.M2, ECL.M], 32) \
        .when([Version.M3, ECL.L], 84) \
        .when([Version.M3, ECL.M], 68) \
        .when([Version.M4, ECL.L], 128) \
        .when([Version.M4, ECL.M], 112) \
        .when([Version.M4, ECL.Q], 80) \
        .value


def get_error_correction_codeword_num(version: Version, ecl: ECL) -> int:
    """
    誤り訂正コード語の総数を取得

    P36 (PDF 39) 表9

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :return: 誤り訂正コード語の総数
    """
    return Case([version, ecl]) \
        .when([Version.M1, ECL.NONE], 2) \
        .when([Version.M2, ECL.L], 5) \
        .when([Version.M2, ECL.M], 6) \
        .when([Version.M3, ECL.L], 6) \
        .when([Version.M3, ECL.M], 8) \
        .when([Version.M4, ECL.L], 8) \
        .when([Version.M4, ECL.M], 10) \
        .when([Version.M4, ECL.Q], 14) \
        .value


def get_symbol_number(version: Version, ecl: ECL) -> int:
    """
    シンボル番号を取得

    P55 (PDF 58) 表13

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :return: シンボル番号
    """
    lst = [
        (Version.M1, ECL.NONE),
        (Version.M2, ECL.L),
        (Version.M2, ECL.M),
        (Version.M3, ECL.L),
        (Version.M3, ECL.M),
        (Version.M4, ECL.L),
        (Version.M4, ECL.M),
        (Version.M4, ECL.Q),
    ]
    return lst.index((version, ecl))


def get_format_information(version: Version, ecl: ECL, mask: Mask) -> BinaryArray:
    """
    形式情報を取得

    P55 (PDF 58) 7.9.2

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :param mask: マスク
    :return: 形式情報のビット列
    """
    symbol_number = get_symbol_number(version, ecl)
    fi = symbol_number << 2 | mask.mask_pattern_value
    fi_poly = PolynomialRing(fi << 10)
    gp = PolynomialRing(0b101_0011_0111)
    bch = (fi_poly % gp).coefficient
    x = concat_arr([bin2arr(fi, 5), bin2arr(bch, 10)])
    m = bin2arr(0b100010001000101, 15)
    return x ^ m


def get_mode_indicator(version: Version, mode: Mode) -> BinaryArray:
    """
    モード指示子を取得

    P21 (PDF 24) 表2

    :param version: 型番
    :param mode: モード
    :return: モード指示子のビット列
    """
    _assert_mode(version, mode)

    mi_len = version.mode_indicator_length
    mi_val = mode.mode_indicator_value
    return bin2arr(mi_val, mi_len)


def get_character_count_indicator_length(version: Version, mode: Mode) -> int:
    """
    文字数指示子のビット数を取得

    P21 (PDF 24) 表3

    :param version: 型番
    :param mode: モード
    :return: 文字数指示子のビット数
    """
    _assert_mode(version, mode)

    cci_len = Case([version, mode]) \
        .when([Version.M1, Mode.Numeric], 3) \
        .when([Version.M2, Mode.Numeric], 4) \
        .when([Version.M2, Mode.AlphaNumeric], 3) \
        .when([Version.M3, Mode.Numeric], 5) \
        .when([Version.M3, Mode.AlphaNumeric], 4) \
        .when([Version.M3, Mode.EightBitByte], 4) \
        .when([Version.M3, Mode.Kanji], 3) \
        .when([Version.M4, Mode.Numeric], 6) \
        .when([Version.M4, Mode.AlphaNumeric], 5) \
        .when([Version.M4, Mode.EightBitByte], 5) \
        .when([Version.M4, Mode.Kanji], 4) \
        .value
    return cci_len


def get_character_count_indicator(version: Version, mode: Mode, character_count: int) -> BinaryArray:
    """
    文字数指示子を取得

    P21 (PDF 24) 表3

    :param version: 型番
    :param mode: モード
    :param character_count: 文字数
    :return: 文字数指示子のビット列
    """
    cci_len = get_character_count_indicator_length(version, mode)
    try:
        return bin2arr(character_count, cci_len)
    except ValueError as err:
        raise OverCapacityError() from err
# endregion


# region 組み合わせの確認
def _check_combination_ve(version: Version, ecl: ECL) -> bool:
    """
    型番と誤り訂正レベルの組み合わせが有効か確認する

    P31 (PDF 34) 表7, P36 (PDF 39) 表9 等

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :return: 組み合わせが有効か？
    """
    if version == Version.M1 and ecl != ECL.NONE:
        # logger.debug(f'M1 supports only error detection')
        return False  # M1は誤り検出のみ

    if ecl == ECL.NONE and version != Version.M1:
        # logger.debug(f'error detection supports only M1')
        return False  # 誤り検出はM1のみ

    if ecl == ECL.Q and version != Version.M4:
        # logger.debug(f'Level Q supports only M4')
        return False  # 誤り訂正レベルQはM4のみ

    return True


def _check_combination_vm(version: Version, mode: Union[Mode, Set[Mode]]) -> bool:
    """
    型番とモードの組み合わせが有効か確認する

    P18 (PDF 21) 7.3, P21 (PDF 24) 表2,3 等

    :param version: 型番
    :param mode: モード
    :return: 組み合わせが有効か？
    """
    modes = {mode} if isinstance(mode, Mode) else mode

    if version == Version.M1 and not modes <= {Mode.Numeric}:
        # logger.debug(f'M1 supports only Numeric mode')
        return False  # M1は数字モードのみ

    if version == Version.M2 and not modes <= {Mode.Numeric, Mode.AlphaNumeric}:
        # logger.debug(f'M2 supports only Numeric mode and AlphaNumeric mode')
        return False  # M2は数字モードと英数字モードのみ

    return True


def check_combination(*, version: Version = None, ecl: ECL = None, mode: Union[Mode, Set[Mode]] = None) -> bool:
    """
    組み合わせが有効か確認する

    :param version: 型番
    :param ecl: 誤り訂正レベル
    :param mode: モード
    :return: 組み合わせが有効か？
    """
    if version is not None and ecl is not None:
        if not _check_combination_ve(version, ecl):
            return False

    if version is not None and mode is not None:
        if not _check_combination_vm(version, mode):
            return False

    return True
# endregion



from enum import Enum


class ErrorCorrectionLevel(Enum):
    """
    誤り訂正レベル

    P34 (PDF 37) 7.5
    """

    NONE = 0
    """誤り検出のみ"""

    L = 1
    """復元能力 約7%"""

    M = 2
    """復元能力 約15%"""

    Q = 3
    """復元能力 約25%"""

    # マイクロQRコードでは使用できないので除外
    # H = 4
    # """復元能力 約30%"""



"""
マスクの行列を生成するプログラム
"""

import itertools
from logging import getLogger
from typing import Tuple, Union

from ..binary import BinaryMatrix, empty_matrix
from ..model import Mask

logger = getLogger(__name__)


def get_mask_matrix(mask: Mask, size: Union[int, Tuple[int, int]]) -> BinaryMatrix:
    """
    マスクを取得

    :param mask: マスクの種類
    :param size: 行列の大きさ
    :return: マスクの行列
    """

    def func(tup):
        i, j = tup

        # データ領域以外はマスクしない
        if i == 0 or j == 0:  # タイミングパターン
            return False
        if i <= 8 and j <= 8:  # 切り出しシンボル・分離パターン・形式情報
            return False

        return mask.function(i, j)  # マスクパターン

    mat = empty_matrix(size)
    h, w = mat.shape
    idx = tuple(zip(*filter(func, itertools.product(range(h), range(w)))))
    mat[idx] = True
    return mat


def calc_mask_score(matrix: BinaryMatrix) -> int:
    """
    マスクの点数を計算

    :param matrix: 点数を計算する行列
    :return: 得点(高い方が良い)
    """
    s1 = sum(matrix[-1, 1:])
    s2 = sum(matrix[1:, -1])
    return min(s1, s2) * 16 + max(s1, s2)


def get_optimal_mask(code: BinaryMatrix) -> Tuple[Mask, BinaryMatrix]:
    """最適なマスクを取得"""

    best_mask: Tuple[Mask, BinaryMatrix] = ...
    best_score = -1

    logger.debug(f'[select optimal mask]')
    for mask in Mask:
        mat_mask = get_mask_matrix(mask, code.shape)
        score = calc_mask_score(mat_mask ^ code)
        logger.debug(f'- {mask}: {score:3}')
        if score > best_score:
            best_mask = mask, mat_mask
            best_score = score

    logger.info(f'selected mask: {best_mask[0]}')
    return best_mask



"""
組み込み例外を詳細に分類するためのファイル
"""


class InvalidPairError(ValueError):
    """不正な組み合わせ"""
    def __init__(self, *args):
        super().__init__(*args)


class InvalidCharacterError(ValueError):
    """不正な文字"""
    def __init__(self, *args):
        super().__init__(*args)


class OverCapacityError(ValueError):
    """容量オーバー"""
    def __init__(self, *args):
        super().__init__(*args)



"""
バイナリ形式のデータを使うための内部モジュール
"""

# それぞれの実体はnumpyのndarrayだが、隠蔽して使用する
# (numpyはこの内部モジュール以外ではimportしないようにする)

from .array import (
    BinaryArray,
    arr2bin,
    bin2arr,
    arr2str,
    concat_arr,
)
from .matrix import (
    BinaryMatrix,
    bin2mat,
    mat2str,
    merge_matrix,
    toggle_matrix,
    empty_matrix,
)



"""
python -m miqro_qr
"""

import argparse
import sys
from logging import getLogger, StreamHandler, Formatter, DEBUG, INFO
from typing import Optional

import mkmqr
from .factory import create_symbol_image
from .model import ErrorCorrectionLevel as ECL, InvalidPairError, InvalidCharacterError, OverCapacityError, set_encoding

handler = StreamHandler()
# handler.setFormatter(Formatter('[{levelname:>8}] {filename:20} L{lineno:3}, {funcName:25} : {message}', style='{'))
handler.setFormatter(Formatter('%(message)s'))
logger = getLogger(mkmqr.__name__)
logger.addHandler(handler)


def main():
    # help_ecl = '誤り訂正レベル'
    # help_path = '画像を保存するパス'
    # help_encoding = 'エンコーディング (bビットバイトモードのみ)'
    # help_show = '画像を表示する'
    # help_debug = 'デバッグ情報を表示'
    # help_text = 'マイクロQRコードにするテキスト'
    help_ecl = 'required error correction level'
    help_path = 'path to save the image'
    help_encoding = 'encoding (8-bit byte mode only)'
    help_show = 'show the image'
    help_debug = 'show debug message'
    help_text = 'text to make micro QR code'

    ecls = {
        'x': ECL.NONE,
        'l': ECL.L,
        'm': ECL.M,
        'q': ECL.Q,
    }

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-e', '--ecl',
        choices=ecls.keys(), default='x', help=help_ecl
    )
    parser.add_argument(
        '-p', '--path',
        help=help_path
    )
    parser.add_argument(
        '--encoding',
        default='shift-jis',
        help=help_encoding
    )
    parser.add_argument(
        '-s', '--show',
        action='store_true',
        help=help_show
    )
    parser.add_argument(
        '-d', '--debug',
        action='count',
        help=help_debug
    )
    parser.add_argument(
        'text',
        help=help_text
    )

    args = parser.parse_args()
    ecl: ECL = ecls[args.ecl]
    path: Optional[str] = args.path
    encoding: str = args.encoding
    show: bool = args.show
    debug: Optional[int] = args.debug
    text: str = args.text

    if not show:
        if path is None:
            path = f'./{text}.png'
        elif path.endswith('/'):
            path = path + f'{text}.png'

    if debug:
        if debug == 1:
            level = INFO
        else:  # debug >= 2
            level = DEBUG
        logger.setLevel(level)
        handler.setLevel(level)

    try:
        logger.info(f'ecl:      {ecl}')
        logger.info(f'path:     {path}')
        logger.info(f'show:     {show}')
        logger.info(f'encoding: {encoding}')
        logger.info(f'debug:    {debug}')
        logger.info(f'text:     {text}')
        logger.info('------------------------------------')

        set_encoding(encoding)
        image = create_symbol_image(text, ecl)

        if path is not None:
            image.save(path)
            logger.info(f'image saved')
        if show:
            image.show(text)
    except OverCapacityError:
        print(f'over capacity', file=sys.stderr)
    except InvalidCharacterError as e:
        print(f'invalid character : {e.args[1]}', file=sys.stderr)
    except InvalidPairError:
        print(f'invalid pair', file=sys.stderr)


if __name__ == '__main__':
    main()
